#include "bulletinfo.h"
#include "cpputil.h"
#include "surface.h"
#include "conf.h"

#include <fstream>
#include <cassert>
#include <boost/functional.hpp>
#include <boost/compose.hpp>
#include <boost/lexical_cast.hpp>

#ifdef HAVE_DIRENT_H
# include <dirent.h>
#endif

void BulletDescription::loadUserInfo(const std::string& userName) {
#ifndef ZAURUS
	std::ifstream is((DAT_DIR "save/"+userName+file_+".dat").c_str());
	if (is.is_open()) {
		is >> come_ >> damage_ >> limit_;
	}
	else {
		come_ = damage_ = limit_ = 0;
	}
#endif
}

void BulletDescription::saveUserInfo(const std::string& userName) const {
#ifndef ZAURUS
	std::ofstream os((DAT_DIR "save/"+userName+file_+".dat").c_str());
	os << come_ << std::endl
	   << damage_ << std::endl
	   << limit_ << std::endl;
#endif
}

void BulletInfo::loadUserInfo(const std::string& userName) {
#ifndef ZAURUS
	userName_ = userName;
	std::for_each(descriptionsIdOrder_.begin(), descriptionsIdOrder_.end(),
				  boost::bind2nd(
					  boost::mem_fun(&BulletDescription::loadUserInfo),
					  userName_));
#endif
}

void BulletInfo::saveUserInfo() const {
#ifndef ZAURUS
	std::for_each(descriptionsIdOrder_.begin(), descriptionsIdOrder_.end(),
				  boost::bind2nd(
					  boost::mem_fun(&BulletDescription::saveUserInfo),
					  userName_));
#endif
}

std::string BulletDescription::xmlFile() const {
	if (userDefined_) {
		return file();
	}
	else {
		return DAT_DIR "bosses.d/" + file() + ".xml";
	}
}

BulletDescription::BulletDescription(const std::string& file,
									 const std::string& title, int id)
	:  id_(id), file_(file), title_(title), userDefined_(true)
#ifdef ZAURUS
	, score_(0)
#endif
{}

BulletDescription::BulletDescription(std::istream& is)
	: userDefined_(false)
#ifdef ZAURUS
	, score_(0)
#endif
{
	static const int MAXBUF = 1000;

	char ln[MAXBUF];
	is.getline(ln, MAXBUF);
	if (ln[0] == '\0') return;
	id_ = boost::lexical_cast<int>(ln);
	is.getline(ln, MAXBUF);
	file_ = ln;
	is.getline(ln, MAXBUF);
	title_ = ln;
	is.getline(ln, MAXBUF);
	description_ = ln;
	is.getline(ln, MAXBUF);
	capture_ = ln;

#ifndef ZAURUS
	if (file_ != "main" && !file_.empty()) {
		std::string bmp("shots/" + file_ + ".bmp");
		if (std::ifstream(bmp.c_str()).is_open()) {
			hasShot_ = true;
			if (Conf::instance()->preloadShots()) {
				shot_.reset(new Surface(bmp));
			}
		}
		else {
			hasShot_ = false;
			std::cerr << "warn: " + bmp + " not found." << std::endl;
		}
	}
#endif
}

BulletInfo::BulletInfo(const std::string& str) {
	std::ifstream is(str.c_str());
	check(is.is_open(), str + " is not found.");

	int i = 0;
	while (!is.eof()) {
		try {
			BulletDescription* desc = new BulletDescription(is);
			if (desc->capture() == "") {
				delete desc;
				break;
			}
			descriptionsDisplayOrder_.push_back(desc);
		}
		catch (...) {
			throw std::runtime_error("boss data file is broken in line "
				+ boost::lexical_cast<std::string>(i));
		}

		++i;
	}

	officialBulletNum_ = i - 1;

	descriptionsIdOrder_.assign(descriptionsDisplayOrder_.begin(),
							   descriptionsDisplayOrder_.end());
	std::sort(descriptionsIdOrder_.begin(),
			  descriptionsIdOrder_.end(),
			  std::mem_fun(&BulletDescription::isIdLess));

#ifdef ZAURUS
	std::ifstream is2(DAT_DIR "save/score.dat");
	if (is2.is_open()) {
		for (size_t j = 0; j < descriptionsIdOrder_.size(); j++) {
			int s;
			if (!(is2 >> s)) break;
			descriptionsIdOrder_[j]->setScore(s);
		}
	}
#endif

#ifdef HAVE_DIRENT_H
	DIR* dir = opendir("user");
	if (dir != 0) {
		int j = 0;
		while (dirent* d = readdir(dir)) {
			std::string name(std::string("user/") + d->d_name);
			if (name.find(".xml") == name.size() - 4) {
				std::string title(std::string("UserDefined")
							  + boost::lexical_cast<std::string>(++j));
				BulletDescription* desc =
					new BulletDescription(name, title, i);
				descriptionsDisplayOrder_.push_back(desc);
				descriptionsIdOrder_.push_back(desc);
				std::cout << name << ": user defined bullet" << std::endl;
				++i;
			}
		}
	}
#endif // HAVE_DIRENT_H

	std::cout << "registered " << i-1 << " bullets." << std::endl;
}

BulletInfo::~BulletInfo() {
	delete_clear(descriptionsIdOrder_);
}

void BulletInfo::saveScore() const {
	std::ofstream os(DAT_DIR "save/score.dat");
	for (size_t i = 0; i < descriptionsIdOrder_.size(); i++) {
		os << descriptionsIdOrder_[i]->score() << std::endl;
	}
}
